#include "engine.h"
#include "posprintf.h"

extern const s8 player[];
extern const s8 swimmer[];
extern const s8 rocket_boat[];

/******************************************************************************/

void insert_node(entity_t** pTreeNode, entity_t* newNode)
{
  entity_t* treeNode = *pTreeNode;

  if (treeNode == 0)
    *pTreeNode = newNode;
  else if (newNode->grid < treeNode->grid) {
    insert_node(&(treeNode->left), newNode);
    if (newNode->parent == 0)
      newNode->parent = treeNode;
  }
  else if (newNode->grid > treeNode->grid) {
    insert_node(&(treeNode->right), newNode);
    if (newNode->parent == 0)
      newNode->parent = treeNode;
  }
  else {
    if (newNode->y > treeNode->y) {
      newNode->parent = treeNode->parent;
      newNode->left = treeNode->left;
      newNode->right = treeNode;
      treeNode->parent = newNode;
      treeNode->left = 0;
      *pTreeNode = newNode;
    }
    else if (treeNode->right != 0 && treeNode->right->grid != newNode->grid) {
      newNode->parent = treeNode;
      newNode->right = treeNode->right;
      treeNode->right->parent = newNode;
      treeNode->right = newNode;
    }
    else {
      insert_node(&(treeNode->right), newNode);
      if (newNode->parent == 0)
        newNode->parent = treeNode;
    }
  }
}

void traverse_binary_tree(entity_t* node, entity_behaviour_t callback)
{
  if (node == 0)
    return;
  traverse_binary_tree(node->left, callback);
  callback(node);
  traverse_binary_tree(node->right, callback);
}

entity_t* search_binary_tree(entity_t* node, u16 grid)
{
  if (node == 0)
    return 0;
  if (grid < node->grid)
    return search_binary_tree(node->left, grid);
  else if (grid > node->grid)
    return search_binary_tree(node->right, grid);
  else
    return node;
}

entity_t* find_min(entity_t* node)
{
  entity_t* current_node = node;
  while (current_node->left)
    current_node = current_node->left;
  return current_node;
}

//Removes the reference to node from node->parent and replaces it with new_node.
void replace_node_in_parent(entity_t* node, entity_t* new_node)
{
  if (node == node->parent->left)
    node->parent->left = new_node;
  else
    node->parent->right = new_node;
  if (new_node)
    new_node->parent = node->parent;
}

void binary_tree_delete(entity_t** root, entity_t* node)
{
  if (node->left && node->right) { // if both children are present
    // get the smallest node that's bigger than *self*
    entity_t* successor = find_min(node->right);
            
    // if *successor* has a child, replace it with that
    // at this point, it can only have a *right_child*
    // if it has no children, *right_child* will be "Null"
    replace_node_in_parent(successor, successor->right);
    successor->left = node->left;
    successor->right = node->right;
    if (node->parent)
      replace_node_in_parent(node, successor);
    else {
      *root = successor;
      successor->parent = 0;
    }
              
  }
  else if (node->left || node->right) {  // if the node has only one child
    if (node->left)
      if (node->parent)
        replace_node_in_parent(node, node->left);
      else {
        *root = node->left;
        node->left->parent = 0;
      }
    else
      if (node->parent)
        replace_node_in_parent(node, node->right);
      else {
        *root = node->right;
        node->right->parent = 0;
      }
  }
  else // this node has no children
    if (node->parent)
      replace_node_in_parent(node, 0);
    else
      *root = 0;

  node->parent = node;
  node->left = 0;
  node->right = 0;
}

void move_node_up(entity_t** root, entity_t* node)
{
  entity_t* tmpNode = node->parent->left;
  if (tmpNode)
    tmpNode->parent = node;

  if (node->parent != *root) {
    if (node->parent == node->parent->parent->left)
      node->parent->parent->left = node;
    else
      node->parent->parent->right = node;
  }
  else
    *root = node;
  
  if (node->left)
    node->left->parent = node->parent;
  if (node->right)
    node->right->parent = node->parent;
  node->parent->left = node->left;
  node->parent->right = node->right;
  node->left = tmpNode;
  node->right = node->parent;
  node->parent = node->parent->parent;
  node->right->parent = node;
}

/*
** Name: engine_entity
** Desc: Creates a new entity and returns a entity pointer.
*/
entity_t* engine_entity(engine_t* engine, int x, int y, int z, entity_behaviour_t behaviour)
{
  int i;
  entity_t* e;
  for (i = 0; i < ENGINE_ENTITY_MAX; i++) {
    e = &engine->entities[i];
    if (e->parent == e)
      break;
  }

  if (e->parent != e)
    return 0;

  CLEARMEM(e);
  
  e->grid = (f2i(y) << 7) + f2i(x);
  e->x = x;
  e->y = y;
  e->z = z;
  e->ontick = behaviour;
  e->life = 127;
  e->engine = engine;

  insert_node(&(engine->root), e);

  return e;
}

int entity_collision(engine_t* engine, entity_t* e, vec2_t v, vec2_t* b)
{
  s16 i, j;
  u16* cellidx;
  vec2_t vb;
  h_cell_t* hc;
  s32 xPmin, xPmax, yPmin, yPmax, xQmin, xQmax, yQmin, yQmax;
  int c = fixcos(f2i(e->t)), s = fixsin(f2i(e->t));
  s32 xPdif, yPdif, cxPdf, sxPdf, cyPdf, syPdf;

  xPdif = ((s8*)e->visual.data)[2];
  yPdif = ((s8*)e->visual.data)[4] >> 1;
  cxPdf = c * xPdif;
  sxPdf = s * xPdif;
  cyPdf = c * yPdif;
  syPdf = s * yPdif;

  //Compute the bounding box coordinates for P.
  if (c > 0)
    if (s > 0)
    {
      xPmin = -( xPmax = f2i(cxPdf + syPdf) );
      yPmin = -( yPmax = f2i(cyPdf + sxPdf) );
    }
    else //s <= 0
    {
      xPmin = -( xPmax = f2i(cxPdf - syPdf) );
      yPmin = -( yPmax = f2i(cyPdf - sxPdf) );
    }
  else //c <= 0
    if (s > 0)
    {
      xPmax = -( xPmin = f2i(cxPdf - syPdf) );
      yPmax = -( yPmin = f2i(cyPdf - sxPdf) );
    }
    else //s <= 0
    {
      xPmax = -( xPmin = f2i(cxPdf + syPdf) );
      yPmax = -( yPmin = f2i(cyPdf + sxPdf) );
    }

  for (i = 0; i < 2; i++)
  {
    vb.x = v.x + b[i].x;
    vb.y = v.y + b[i].y;
    hc = (h_cell_t*)&huntercells[engine->grid->cells[f2i(vb.y)][f2i(vb.x)]];
    //White border
    if (hc->color == 9)
      return 1;
    //Land
    if (e->visual.data == &rocket_boat)
    {
      switch (hc->style)
      {
        case 4: if (hc->color != 1) return 1; break;
        case 0: if ((vb.x & 0xff) > (vb.y & 0xff)) return 1; break;
        case 1: if ((vb.x & 0xff) < 255-(vb.y & 0xff)) return 1; break;
        case 2: if ((vb.x & 0xff) < (vb.y & 0xff)) return 1; break;
        case 3: if ((vb.x & 0xff) > 255-(vb.y & 0xff)) return 1; break;
      }
    }
    //Water
    else if (e->visual.data != &player && e->visual.data != &swimmer)
    {
      if (hc->color == 1)
        return 1;
      switch (hc->style)
      {
        case 4: break;
        case 0: if ((vb.x & 0xff) < (vb.y & 0xff)) return 1; break;
        case 1: if ((vb.x & 0xff) > 255-(vb.y & 0xff)) return 1; break;
        case 2: if ((vb.x & 0xff) > (vb.y & 0xff)) return 1; break;
        case 3: if ((vb.x & 0xff) < 255-(vb.y & 0xff)) return 1; break;
      }
    }
  }

  cellidx = &engine->grid->cells[f2i(v.y)+1][f2i(v.x)-1];

  for (i = -256; i < 512; i+=256, cellidx-=131)
  {
    for (j = -256; j < 512; j+=256, cellidx++)
    {
      hc = (h_cell_t*)&huntercells[*cellidx];

    if (hc->entityID >= 0)
    {
      rom_entity_t* prent = (rom_entity_t*)&hunterentities[hc->entityID];

      do
      {
        if (engine->building_strength[prent - hunterentities] == 0)
          continue;

        //xQmin = (vb.x & 0xffffff00) + prent->x - ((s8*)prent->model)[2] - v.x - f2i(syPdf);
        xQmin = prent->x - ((s8*)prent->model)[2] - (v.x & 0xff) + j - f2i(syPdf);
        xQmax = xQmin + (((s8*)prent->model)[2] << 1);
        //yQmin = (vb.y & 0xffffff00) + prent->y - ((s8*)prent->model)[4] - v.y - f2i(cyPdf);
        yQmin = prent->y - ((s8*)prent->model)[4] - (v.y & 0xff) - i - f2i(cyPdf);
        yQmax = yQmin + (((s8*)prent->model)[4] << 1);

        if ( xPmin > xQmax || xQmin > xPmax || yPmin > yQmax || yQmin > yPmax )
          continue;

        if (c > 0)
        {
          if (s > 0)
          {
            if (!(f2i(c*xQmax + s*yQmax) < -xPdif ||
                  f2i(c*xQmin + s*yQmin) >  xPdif ||
                  f2i(c*yQmax - s*xQmin) < -yPdif ||
                  f2i(c*yQmin - s*xQmax) >  yPdif)) {
              if (e->onBGhit)
                e->onBGhit(e, prent - hunterentities);
              return 1;
            }
          }
          else //s <= 0
            if (!(f2i(c*xQmax + s*yQmin) < -xPdif ||
                  f2i(c*xQmin + s*yQmax) >  xPdif ||
                  f2i(c*yQmax - s*xQmax) < -yPdif ||
                  f2i(c*yQmin - s*xQmin) >  yPdif)) {
              if (e->onBGhit)
                e->onBGhit(e, prent - hunterentities);
              return 1;
            }
        }
        else //c <= 0
        {
          if (s > 0)
          {
            if (!(f2i(c*xQmin + s*yQmax) < -xPdif ||
                  f2i(c*xQmax + s*yQmin) >  xPdif ||
                  f2i(c*yQmin - s*xQmin) < -yPdif ||
                  f2i(c*yQmax - s*xQmax) >  yPdif)) {
              if (e->onBGhit)
                e->onBGhit(e, prent - hunterentities);
              return 1;
            }
          }
          else //s <= 0
            if (!(f2i(c*xQmin + s*yQmin) < -xPdif ||
                  f2i(c*xQmax + s*yQmax) >  xPdif ||
                  f2i(c*yQmin - s*xQmax) < -yPdif ||
                  f2i(c*yQmax - s*xQmin) >  yPdif)) {
              if (e->onBGhit)
                e->onBGhit(e, prent - hunterentities);
              return 1;
            }
        }

      } while ((prent++)->more > 0);
    }
    }
  }
  e->x = v.x;
  e->y = v.y;
  return 0;
}

/*
** Name: entity_motion
** Desc: Simple map to entity motion & collision.
*/
void entity_motion(entity_t* e)
{
  engine_t* engine = e->engine;

  int dy, dx, stepx, stepy;
  vec2_t old, v, target, front, side;
  vec2_t bumper[2];

  old.x = v.x = e->x;
  old.y = v.y = e->y;
  dy = e->yy;
  dx = e->xx;
  target.x = v.x + dx;
  target.y = v.y + dy;
  front.x = f2i(fixsin(f2i(e->t)) * ((s8*)e->visual.data)[4]);
  front.y = f2i(fixcos(f2i(e->t)) * ((s8*)e->visual.data)[4]);
  side.x = f2i(fixsin(f2i(e->t)+512) * ((s8*)e->visual.data)[2]);
  side.y = f2i(fixcos(f2i(e->t)+512) * ((s8*)e->visual.data)[2]);
  bumper[0].x = front.x + side.x;
  bumper[0].y = front.y + side.y;
  bumper[1].x = front.x - side.x;
  bumper[1].y = front.y - side.y;

  if (dy < 0) { dy = -dy;  stepy = -1; } else { stepy = 1; }
  if (dx < 0) { dx = -dx;  stepx = -1; } else { stepx = 1; }
  dy <<= 1;
  dx <<= 1;

  if (dx > dy) {
    int fraction = dy - (dx >> 1);
    while (v.x != target.x) {
      if (fraction >= 0) {
        v.y += stepy;
        fraction -= dx;
      }
      v.x += stepx;
      fraction += dy;
      if (entity_collision(engine, e, v, bumper))
        break;
    }
  } else {
    int fraction = dx - (dy >> 1);
    while (v.y != target.y) {
      if (fraction >= 0) {
        v.x += stepx;
        fraction -= dy;
      }
      v.y += stepy;
      fraction += dx;
      if (entity_collision(engine, e, v, bumper))
        break;
    }
  }

  e->z += e->zz;
  e->r += e->rr; e->t += e->tt; e->p += e->pp;

  if (f2i(old.x) != f2i(e->x) || f2i(old.y) != f2i(e->y)) {
    e->grid = (f2i(e->y) << 7) + f2i(e->x);
    binary_tree_delete(&(engine->root), e);
    e->parent = 0;
    insert_node(&(engine->root), e);
  }
  else if (e->parent && e->parent->grid == e->grid && e->y > e->parent->y)
    move_node_up(&(engine->root), e);
}

/******************************************************************************/

void engine_clear_entities(engine_t* engine)
{
  int i;
  engine->nentities = 0;
  engine->root = 0;

  for (i = 0; i < ENGINE_ENTITY_MAX; i++)
    engine->entities[i].parent = &engine->entities[i];

  engine->camera = engine_entity(engine,
    i2fdiv2(ENGINE_MAP_WIDTH),
    i2fdiv2(ENGINE_MAP_HEIGHT),
    i2f(0), 0);
}

void engine_load_map(engine_t* engine, rom_map_t* src, h_map_t* hsrc)
{
  int x, y, i;

  engine->curr_map = src;
  engine->grid = hsrc;
  if (!engine->camera) engine_clear_entities(engine);
  
  for (y = 0; y < ENGINE_MAP_HEIGHT; y++)
  {
    for (x = 0; x < ENGINE_MAP_WIDTH; x++)
    {
      if (engine->curr_map->cells[y][x].ent == 1)
      {
        engine->camera->x = i2f(x) + FIXHALF;
        engine->camera->y = i2f(y) + FIXHALF;
      }
    }
  }
  engine->camera->t = 0;
  engine->camera->r = 0;
  engine->camera->p = 0;

  for (i = 0; i < 100; i++)
    engine->building_strength[i] = 100;
}

/******************************************************************************/

/*
** Name: isqrt
** Desc: Integer square root. Take the square root of an integer.
*/
int isqrt(int value)
{
  int root = 0;

#define STEP(shift) \
    if((0x40000000 >> shift) + root <= value)          \
    {                                                   \
        value -= (0x40000000 >> shift) + root;          \
        root = (root >> 1) | (0x40000000 >> shift);     \
    }                                                   \
    else                                                \
    {                                                   \
        root >>= 1;                                     \
    }

  STEP( 0); STEP( 2); STEP( 4); STEP( 6);
  STEP( 8); STEP(10); STEP(12); STEP(14);
  STEP(16); STEP(18); STEP(20); STEP(22);
  STEP(24); STEP(26); STEP(28); STEP(30);

  // round to the nearest integer, cuts max error in half

  if (root < value) root++;

  return root;
}

/******************************************************************************/

/*
** Name: engine_init
** Desc: Setup a default map and position the camera.
*/
void engine_init(
  engine_t* engine,
  framebuffer_t* front,
  framebuffer_t* back)
{
  CLEARMEM(engine);
}

void engine_tick(engine_t* engine)
{
  entity_t* e1;

  engine->tick++;

  for (e1 = engine->entities; e1 != &engine->entities[ENGINE_ENTITY_MAX]; e1++)
    if (e1->parent != e1 && e1->ontick)
      e1->ontick(e1);

  for (e1 = engine->entities; e1 != &engine->entities[ENGINE_ENTITY_MAX]; e1++)
  {
    if (e1->parent == e1)
      continue;

    entity_motion(e1);

    if (e1->radius)
    {
      entity_t* e2;
      
      for (e2 = e1+1; e2 != &engine->entities[ENGINE_ENTITY_MAX]; e2++)
      {
        if (e2->radius)
        {
          int size = e1->radius + e2->radius;
          int x = e1->x - e2->x;
          int y = e1->y - e2->y;
          int z = e1->z - e2->z;

          if (
            x < size && x > -size &&
            y < size && y > -size &&
            z < size && z > -size )
          {
            if (e1->onhit) e1->onhit(e1, e2);
            if (e2->onhit) e2->onhit(e2, e1);
          }
        }
      }
    }
  }
  camera_behaviour(engine->camera);
  engine->keyboard._pad = engine->keyboard.pad;
}

/******************************************************************************/

void* engine_memcpy(void* d, void* s, unsigned n)
{
  while (n--) ((u8*)d)[n] = ((u8*)s)[n];
  return d;
}

void* engine_memset(void *s, int c, unsigned n)
{
  while (n--) ((u8*)s)[n] = c;
  return s;
}

/******************************************************************************/
